import numpy as np
from itertools import product

from ..base import BaseBinaryEncoder
from ...helper import batching


class QuantizerEncoder(BaseBinaryEncoder):
    batch_size = 2048

    def __init__(self, dim_per_byte: int, cluster_per_byte: int = 255,
                 upper_bound: int = 10000,
                 lower_bound: int = -10000,
                 partition_method: str = 'average',
                 *args, **kwargs):
        super().__init__(*args, **kwargs)
        assert 1 < cluster_per_byte <= 255, 'cluster number should >1 and <= 255 (0 is reserved for NOP)'
        self.dim_per_byte = dim_per_byte
        self.num_clusters = cluster_per_byte
        self.upper_bound = upper_bound
        self.lower_bound = lower_bound
        self.partition_method = partition_method
        self.centroids = self._get_centroids()

    def _get_centroids(self):
        """
        calculate centroids for quantizer
        two kinds of divide methods are supported now: average, random
        average: split the space averagely and centroids of clusters lie on the corner of sub-space
        random: randomly pick points and treat them as centroids of clusters
        Variable Explaination:
            num_sample_per_dim: number of points to be sample on each dimension
        """

        if self.upper_bound < self.lower_bound:
            raise ValueError("upper bound is smaller than lower bound")

        centroids = []
        num_sample_per_dim = np.ceil(pow(self.num_clusters, 1 / self.dim_per_byte)).astype(np.uint8)
        if self.partition_method == 'average':
            axis_point = np.linspace(self.lower_bound, self.upper_bound, num=num_sample_per_dim + 1,
                                     endpoint=False, retstep=False, dtype=None)[1:]
            coordinates = np.tile(axis_point, (self.dim_per_byte, 1))
        elif self.partition_method == 'random':
            coordinates = np.random.uniform(self.lower_bound, self.upper_bound,
                                            size=[self.dim_per_byte, num_sample_per_dim])
        else:
            raise NotImplementedError

        for item in product(*coordinates):
            centroids.append(list(item))
        return centroids[:self.num_clusters]

    @batching
    def encode(self, vecs: np.ndarray, *args, **kwargs) -> np.ndarray:
        self._check_bound(vecs)
        num_bytes = self._get_num_bytes(vecs)
        x = np.reshape(vecs, [vecs.shape[0], num_bytes, 1, self.dim_per_byte])
        x = np.sum(np.square(x - self.centroids), -1)
        # start from 1
        x = np.argmax(-x, 2) + 1

        return np.array(x, dtype=np.uint8)

    def _get_num_bytes(self, vecs: np.ndarray):
        num_dim = vecs.shape[1]
        assert num_dim % self.dim_per_byte == 0 and num_dim >= (num_dim % self.dim_per_byte), \
            'input dimension (=%d) should be divided by dim_per_byte (=%d)!' % (
                num_dim, self.dim_per_byte)
        return int(num_dim / self.dim_per_byte)

    @staticmethod
    def _get_max_min_value(vecs):
        return np.amax(vecs, axis=None), np.amin(vecs, axis=None)

    def _check_bound(self, vecs):
        max_value, min_value = self._get_max_min_value(vecs)
        if self.upper_bound < max_value:
            raise Warning("upper bound (=%.3f) is smaller than max value of input data (=%.3f), you should choose"
                          "a bigger value for upper bound" % (self.upper_bound, max_value))
        if self.lower_bound > min_value:
            raise Warning("lower bound (=%.3f) is bigger than min value of input data (=%.3f), you should choose"
                          "a smaller value for lower bound" % (self.lower_bound, min_value))
        if (self.upper_bound - self.lower_bound) >= 10 * (max_value - min_value):
            raise Warning("(upper bound - lower_bound) (=%.3f) is 10 times larger than (max value - min value) "
                          "(=%.3f) of data, maybe you should choose a suitable bound" %
                          ((self.upper_bound - self.lower_bound), (max_value - min_value)))



import ctypes
import os
import random
from typing import List, Iterator, Tuple
from typing import Optional

import numpy as np
import zmq
from termcolor import colored

from . import gnes_pb2
from ..helper import batch_iterator, default_logger

__all__ = ['RequestGenerator', 'send_message', 'recv_message',
           'blob2array', 'array2blob', 'gnes_pb2', 'add_route', 'add_version']


class RequestGenerator:
    @staticmethod
    def index(data: Iterator[bytes], batch_size: int = 0, doc_type: int = gnes_pb2.Document.TEXT,
              doc_id_start: int = 0, request_id_start: int = 0,
              random_doc_id: bool = False,
              *args, **kwargs):

        for pi in batch_iterator(data, batch_size):
            req = gnes_pb2.Request()
            req.request_id = request_id_start
            for raw_bytes in pi:
                d = req.index.docs.add()
                d.doc_id = doc_id_start if not random_doc_id else random.randint(0, ctypes.c_uint(-1).value)
                d.raw_bytes = raw_bytes
                d.weight = 1.0
                d.doc_type = doc_type
                doc_id_start += 1
            yield req
            request_id_start += 1

    @staticmethod
    def train(data: Iterator[bytes], batch_size: int = 0, doc_type: int = gnes_pb2.Document.TEXT,
              doc_id_start: int = 0, request_id_start: int = 0,
              random_doc_id: bool = False,
              *args, **kwargs):
        for pi in batch_iterator(data, batch_size):
            req = gnes_pb2.Request()
            req.request_id = request_id_start
            for raw_bytes in pi:
                d = req.train.docs.add()
                d.doc_id = doc_id_start if not random_doc_id else random.randint(0, ctypes.c_uint(-1).value)
                d.raw_bytes = raw_bytes
                d.doc_type = doc_type
                if not random_doc_id:
                    doc_id_start += 1
            yield req
            request_id_start += 1
        req = gnes_pb2.Request()
        req.request_id = request_id_start
        req.train.flush = True
        yield req
        request_id_start += 1

    @staticmethod
    def query(query: bytes, top_k: int, doc_type: int = gnes_pb2.Document.TEXT, request_id_start: int = 0, *args,
              **kwargs):
        if top_k <= 0:
            raise ValueError('"top_k: %d" is not a valid number' % top_k)

        req = gnes_pb2.Request()
        req.request_id = request_id_start
        req.search.query.raw_bytes = query
        req.search.query.doc_type = doc_type
        req.search.top_k = top_k
        yield req


def blob2array(blob: 'gnes_pb2.NdArray') -> np.ndarray:
    """
    Convert a blob proto to an array.
    """
    x = np.frombuffer(blob.data, dtype=blob.dtype).copy()
    return x.reshape(blob.shape)


def array2blob(x: np.ndarray) -> 'gnes_pb2.NdArray':
    """Converts a N-dimensional array to blob proto.
    """
    blob = gnes_pb2.NdArray()
    blob.data = x.tobytes()
    blob.shape.extend(list(x.shape))
    blob.dtype = x.dtype.name
    return blob


def router2str(m: 'gnes_pb2.Message') -> str:
    route_str = [r.service for r in m.envelope.routes]
    return colored('▸', 'green').join(route_str)


def add_route(evlp: 'gnes_pb2.Envelope', name: str, identity: str):
    r = evlp.routes.add()
    r.service = name
    r.start_time.GetCurrentTime()
    r.service_identity = identity


def add_version(evlp: 'gnes_pb2.Envelope'):
    from .. import __version__, __proto_version__
    evlp.gnes_version = __version__
    evlp.proto_version = __proto_version__
    evlp.vcs_version = os.environ.get('GNES_VCS_VERSION', '')


def merge_routes(msg: 'gnes_pb2.Message', prev_msgs: List['gnes_pb2.Message']):
    # take unique routes by service identity
    routes = {(r.service + r.service_identity): r for m in prev_msgs for r in m.envelope.routes}
    msg.envelope.ClearField('routes')
    msg.envelope.routes.extend(sorted(routes.values(), key=lambda x: (x.start_time.seconds, x.start_time.nanos)))


def check_msg_version(msg: 'gnes_pb2.Message'):
    from .. import __version__, __proto_version__
    if hasattr(msg.envelope, 'gnes_version'):
        if not msg.envelope.gnes_version:
            # only happen in unittest
            default_logger.warning('incoming message contains empty "gnes_version", '
                                   'you may ignore it in debug/unittest mode. '
                                   'otherwise please check if frontend service set correct version')
        elif __version__ != msg.envelope.gnes_version:
            raise AttributeError('mismatched GNES version! '
                                 'incoming message has GNES version %s, whereas local GNES version %s' % (
                                     msg.envelope.gnes_version, __version__))

    if hasattr(msg.envelope, 'proto_version'):
        if not msg.envelope.proto_version:
            # only happen in unittest
            default_logger.warning('incoming message contains empty "proto_version", '
                                   'you may ignore it in debug/unittest mode. '
                                   'otherwise please check if frontend service set correct version')
        elif __proto_version__ != msg.envelope.proto_version:
            raise AttributeError('mismatched protobuf version! '
                                 'incoming message has protobuf version %s, whereas local protobuf version %s' % (
                                     msg.envelope.proto_version, __proto_version__))

    if hasattr(msg.envelope, 'vcs_version'):
        if not msg.envelope.vcs_version or not os.environ.get('GNES_VCS_VERSION'):
            default_logger.warning('incoming message contains empty "vcs_version", '
                                   'you may ignore it in debug/unittest mode, '
                                   'or if you run gnes OUTSIDE docker container where GNES_VCS_VERSION is unset'
                                   'otherwise please check if frontend service set correct version')
        elif os.environ.get('GNES_VCS_VERSION') != msg.envelope.vcs_version:
            raise AttributeError('mismatched vcs version! '
                                 'incoming message has vcs_version %s, whereas local environment vcs_version is %s' % (
                                     msg.envelope.vcs_version, os.environ.get('GNES_VCS_VERSION')))

    if not hasattr(msg.envelope, 'proto_version') and not hasattr(msg.envelope, 'gnes_version'):
        raise AttributeError('version_check=True locally, '
                             'but incoming message contains no version info in its envelope. '
                             'the message is probably sent from a very outdated GNES version')


def extract_bytes_from_msg(msg: 'gnes_pb2.Message') -> Tuple:
    doc_bytes = []
    chunk_bytes = []
    doc_byte_type = b''
    chunk_byte_type = b''

    docs = msg.request.train.docs or msg.request.index.docs or [msg.request.search.query]
    # for train request
    for d in docs:
        # oneof raw_data {
        #     string raw_text = 5;
        #       NdArray raw_image = 6;
        #       NdArray raw_video = 7;
        #       bytes raw_bytes = 8; // for other types
        # }
        dtype = d.WhichOneof('raw_data') or ''
        doc_byte_type = dtype.encode()
        if dtype == 'raw_bytes':
            doc_bytes.append(d.raw_bytes)
            d.ClearField('raw_bytes')
        elif dtype == 'raw_image':
            doc_bytes.append(d.raw_image.data)
            d.raw_image.ClearField('data')
        elif dtype == 'raw_video':
            doc_bytes.append(d.raw_video.data)
            d.raw_video.ClearField('data')
        elif dtype == 'raw_text':
            doc_bytes.append(d.raw_text.encode())
            d.ClearField('raw_text')

        for c in d.chunks:
            # oneof content {
            # string text = 2;
            # NdArray blob = 3;
            # bytes raw = 7;
            # }
            chunk_bytes.append(c.embedding.data)
            c.embedding.ClearField('data')

            ctype = c.WhichOneof('content') or ''
            chunk_byte_type = ctype.encode()
            if ctype == 'raw':
                chunk_bytes.append(c.raw)
                c.ClearField('raw')
            elif ctype == 'blob':
                chunk_bytes.append(c.blob.data)
                c.blob.ClearField('data')
            elif ctype == 'text':
                chunk_bytes.append(c.text.encode())
                c.ClearField('text')

    return doc_bytes, doc_byte_type, chunk_bytes, chunk_byte_type


def fill_raw_bytes_to_msg(msg: 'gnes_pb2.Message', msg_data: List[bytes]):
    doc_byte_type = msg_data[2].decode()
    chunk_byte_type = msg_data[3].decode()
    doc_bytes_len = int(msg_data[4])
    chunk_bytes_len = int(msg_data[5])

    doc_bytes = msg_data[6:(6 + doc_bytes_len)]
    chunk_bytes = msg_data[(6 + doc_bytes_len):]

    if len(chunk_bytes) != chunk_bytes_len:
        raise ValueError('"chunk_bytes_len"=%d in message, but the actual length is %d' % (
            chunk_bytes_len, len(chunk_bytes)))

    c_idx = 0
    d_idx = 0
    docs = msg.request.train.docs or msg.request.index.docs or [msg.request.search.query]
    for d in docs:
        if doc_bytes and doc_bytes[d_idx]:
            if doc_byte_type == 'raw_bytes':
                d.raw_bytes = doc_bytes[d_idx]
                d_idx += 1
            elif doc_byte_type == 'raw_image':
                d.raw_image.data = doc_bytes[d_idx]
                d_idx += 1
            elif doc_byte_type == 'raw_video':
                d.raw_video.data = doc_bytes[d_idx]
                d_idx += 1
            elif doc_byte_type == 'raw_text':
                d.raw_text = doc_bytes[d_idx].decode()
                d_idx += 1

        for c in d.chunks:
            if chunk_bytes and chunk_bytes[c_idx]:
                c.embedding.data = chunk_bytes[c_idx]
            c_idx += 1

            if chunk_byte_type == 'raw':
                c.raw = chunk_bytes[c_idx]
                c_idx += 1
            elif chunk_byte_type == 'blob':
                c.blob.data = chunk_bytes[c_idx]
                c_idx += 1
            elif chunk_byte_type == 'text':
                c.text = chunk_bytes[c_idx].decode()
                c_idx += 1


def send_message(sock: 'zmq.Socket', msg: 'gnes_pb2.Message', timeout: int = -1,
                 squeeze_pb: bool = False, **kwargs) -> None:
    try:
        if timeout > 0:
            sock.setsockopt(zmq.SNDTIMEO, timeout)
        else:
            sock.setsockopt(zmq.SNDTIMEO, -1)

        if not squeeze_pb:
            sock.send_multipart([msg.envelope.client_id.encode(), msg.SerializeToString()])
        else:
            doc_bytes, doc_byte_type, chunk_bytes, chunk_byte_type = extract_bytes_from_msg(msg)
            # now raw_bytes are removed from message, hoping for faster de/serialization
            sock.send_multipart(
                [msg.envelope.client_id.encode(),  # 0
                 msg.SerializeToString(),  # 1
                 doc_byte_type, chunk_byte_type,  # 2, 3
                 b'%d' % len(doc_bytes), b'%d' % len(chunk_bytes),  # 4, 5
                 *doc_bytes, *chunk_bytes])  # 6, 7
    except zmq.error.Again:
        raise TimeoutError(
            'cannot send message to sock %s after timeout=%dms, please check the following:'
            'is the server still online? is the network broken? are "port" correct? ' % (
                sock, timeout))
    except Exception as ex:
        raise ex
    finally:
        sock.setsockopt(zmq.SNDTIMEO, -1)


def recv_message(sock: 'zmq.Socket', timeout: int = -1, check_version: bool = False, **kwargs) -> Optional[
    'gnes_pb2.Message']:
    try:
        if timeout > 0:
            sock.setsockopt(zmq.RCVTIMEO, timeout)
        else:
            sock.setsockopt(zmq.RCVTIMEO, -1)

        msg = gnes_pb2.Message()
        msg_data = sock.recv_multipart()
        msg.ParseFromString(msg_data[1])
        if check_version:
            check_msg_version(msg)

        # now we have a barebone msg, we need to fill in data
        if len(msg_data) > 2:
            fill_raw_bytes_to_msg(msg, msg_data)
        return msg

    except zmq.error.Again:
        raise TimeoutError(
            'no response from sock %s after timeout=%dms, please check the following:'
            'is the server still online? is the network broken? are "port" correct? ' % (
                sock, timeout))
    except Exception as ex:
        raise ex
    finally:
        sock.setsockopt(zmq.RCVTIMEO, -1)



from typing import List

from google.protobuf.json_format import MessageToJson, Parse

from ..base import BaseDocIndexer as BDI
from ...proto import gnes_pb2


class DictIndexer(BDI):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._content = {}

    @BDI.update_counter
    def add(self, keys: List[int], docs: List['gnes_pb2.Document'], *args, **kwargs):
        self._content.update({k: MessageToJson(d) for (k, d) in zip(keys, docs)})

    def query(self, keys: List[int], *args, **kwargs) -> List['gnes_pb2.Document']:
        return [Parse(self._content[k], gnes_pb2.Document()) for k in keys]



from ..base import BaseVideoPreprocessor
from ..io_utils import video as video_util
from ...proto import array2blob


class VideoDecoderPreprocessor(BaseVideoPreprocessor):
    store_args_kwargs = True

    def __init__(self,
                 frame_rate: int = 10,
                 frame_size: str = None,
                 vframes: int = -1,
                 drop_raw_data: bool = False,
                 chunk_spliter: str = None,
                 *args,
                 **kwargs):
        super().__init__(*args, **kwargs)
        self.frame_rate = frame_rate
        self.frame_size = frame_size
        self.vframes = vframes
        self.drop_raw_data = drop_raw_data
        self.chunk_spliter = chunk_spliter

    def apply(self, doc: 'gnes_pb2.Document') -> None:
        super().apply(doc)
        if doc.WhichOneof('raw_data'):
            video_frames = []
            # raw_type = type(getattr(doc, doc.WhichOneof('raw_data')))
            if doc.raw_bytes:
                video_frames = video_util.capture_frames(
                    input_data=doc.raw_bytes,
                    scale=self.frame_size,
                    fps=self.frame_rate,
                    vframes=self.vframes)
                if not self.drop_raw_data:
                    doc.raw_video.CopyFrom(array2blob(video_frames))
            else:
                self.logger.error('the document "raw_bytes" is empty!')

            if self.chunk_spliter == 'base':
                for i, frame in enumerate(video_frames):
                    c = doc.chunks.add()
                    c.doc_id = doc.doc_id
                    c.blob.CopyFrom(array2blob(frame))
                    c.offset = i
                    c.weight = 1.0
            elif self.chunk_spliter == 'none':
                pass
            elif self.chunk_spliter == 'shot':
                raise NotImplementedError
            else:
                chunk = doc.chunks.add()
                chunk.doc_id = doc.doc_id
                chunk.blob.CopyFrom(array2blob(video_frames))
                chunk.offset = 0
                chunk.weight = 1.0

        else:
            self.logger.error('bad document: "raw_data" is empty!')



import numpy as np

from ..base import BaseAudioPreprocessor
from ...proto import array2blob, blob2array


class VggishPreprocessor(BaseAudioPreprocessor):

    def __init__(self, num_frames: int = 96,
                 num_bands: int = 64,
                 sample_rate: int = 16000,
                 log_offset: float = 0.01,
                 example_window_seconds: float = 0.96,
                 example_hop_seconds: float = 0.96,
                 stft_window_length_seconds: float = 0.025,
                 stft_hop_length_seconds: float = 0.01,
                 mel_min_hz: int = 125,
                 mel_max_hz: int = 7500,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.num_frames = num_frames
        self.num_bands = num_bands
        self.sample_rate = sample_rate
        self.log_offset = log_offset
        self.example_window_seconds = example_window_seconds
        self.example_hop_seconds = example_hop_seconds
        self.stft_window_length_seconds = stft_window_length_seconds
        self.stft_hop_length_seconds = stft_hop_length_seconds
        self.mel_min_hz = mel_min_hz
        self.mel_max_hz = mel_max_hz
        self.num_mel_binds = num_bands

    def apply(self, doc: 'gnes_pb2.Document') -> None:
        super().apply(doc)

        if doc.raw_bytes:
            for chunks in doc.chunks:
                chunks.blob.CopyFrom(array2blob(np.array(self.waveform_to_examples(blob2array(chunks.blob),
                                                                                   sample_rate=self.sample_rate),
                                                         dtype=np.float32)))
        else:
            self.logger.error('bad document: "raw_bytes" is empty!')

    def waveform_to_examples(self, data, sample_rate):
        """Converts audio waveform into an array of examples for VGGish.

        Args:
          data: np.array of either one dimension (mono) or two dimensions
            (multi-channel, with the outer dimension representing channels).
            Each sample is generally expected to lie in the range [-1.0, +1.0],
            although this is not required.
          sample_rate: Sample rate of data.

        Returns:
          3-D np.array of shape [num_examples, num_frames, num_bands] which represents
          a sequence of examples, each of which contains a patch of log mel
          spectrogram, covering num_frames frames of audio and num_bands mel frequency
          bands, where the frame length is vggish_params.STFT_HOP_LENGTH_SECONDS.
        """
        from .vggish_example_helper import mel_features
        import resampy

        # Convert to mono.
        print(type(data))
        if len(data.shape) > 1:
            data = np.mean(data, axis=1)
        # Resample to the rate assumed by VGGish.
        if sample_rate != self.sample_rate:
            data = resampy.resample(data, sample_rate, self.sample_rate)

        # Compute log mel spectrogram features.
        log_mel = mel_features.log_mel_spectrogram(
            data,
            audio_sample_rate=self.sample_rate,
            log_offset=self.log_offset,
            window_length_secs=self.stft_window_length_seconds,
            hop_length_secs=self.stft_hop_length_seconds,
            num_mel_bins=self.num_mel_binds,
            lower_edge_hertz=self.mel_min_hz,
            upper_edge_hertz=self.mel_max_hz)

        # Frame features into examples.
        features_sample_rate = 1.0 / self.stft_hop_length_seconds
        example_window_length = int(round(
            self.example_window_seconds * features_sample_rate))
        example_hop_length = int(round(
            self.example_hop_seconds * features_sample_rate))
        log_mel_examples = mel_features.frame(
            log_mel,
            window_length=example_window_length,
            hop_length=example_hop_length)
        return log_mel_examples



import unittest
from gnes.proto import gnes_pb2, array2blob, blob2array
from gnes.preprocessor.video.frame_select import FrameSelectPreprocessor
import numpy as np
import copy


class TestFrameSelector(unittest.TestCase):
    def setUp(self) -> None:
        self.doc = gnes_pb2.Document()

        c1 = self.doc.chunks.add()
        c1.blob.CopyFrom(array2blob(np.array([[1,2,3], [2,3,4]])))

        c2 = self.doc.chunks.add()
        c2.blob.CopyFrom(array2blob(np.array([[1,2,3], [2,3,4], [1,2,3]])))

        c3 = self.doc.chunks.add()
        c3.blob.CopyFrom(array2blob(np.array([[1,2,3], [2,3,4], [1,2,3], [2,3,4]])))

    def test_emtpy_document(self):
        frame_selector = FrameSelectPreprocessor(sframes=-1)
        frame_selector.apply(gnes_pb2.Document())

    def test_big_sframe(self):
        doc = copy.deepcopy(self.doc)
        frame_selector = FrameSelectPreprocessor(sframes=100)
        frame_selector.apply(doc)

    def test_get_frames(self):
        doc = copy.deepcopy(self.doc)
        frame_selector = FrameSelectPreprocessor(sframes=3)
        frame_selector.apply(doc)
        for idx, chunk in enumerate(doc.chunks):
            if idx == 0:
                self.assertEqual(blob2array(chunk.blob).shape[0], 2)
            else:
                self.assertEqual(blob2array(chunk.blob).shape[0], 3)

    def test_get_one_frame(self):
        doc = copy.deepcopy(self.doc)
        frame_selector = FrameSelectPreprocessor(sframes=1)
        frame_selector.apply(doc)
        for chunk in doc.chunks:
            self.assertEqual(blob2array(chunk.blob).shape[0], 1)







import json
from typing import Sequence

import numpy as np

from ..base import TrainableBase
from ..proto import gnes_pb2


def get_unary_score(value: float, **kwargs):
    score = gnes_pb2.Response.QueryResponse.ScoredResult.Score()
    score.value = value
    score.explained = json.dumps(
        dict(value=float(value),
             **kwargs))
    return score


class BaseScoreFn(TrainableBase):
    """Base score function. A score function must implement __call__ method"""

    warn_unnamed = False

    def __init__(self, context=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._context = context

    def __call__(self, *args, **kwargs) -> 'gnes_pb2.Response.QueryResponse.ScoredResult.Score':
        raise NotImplementedError

    def new_score(self, *, operands: Sequence['gnes_pb2.Response.QueryResponse.ScoredResult.Score'] = (), **kwargs):
        if not self.__doc__:
            raise NotImplementedError('%s dont have docstring. For the sake of interpretability, '
                                      'please write docstring for this class')
        return get_unary_score(name=self.__class__.__name__,
                               docstring=' '.join(self.__doc__.split()).strip(),
                               operands=[json.loads(s.explained) for s in operands],
                               **kwargs)


class CombinedScoreFn(BaseScoreFn):
    """Combine multiple scores into one score, defaults to 'multiply'"""

    def __init__(self, score_mode: str = 'multiply', *args, **kwargs):
        """
        :param score_mode: specifies how the computed scores are combined
        """
        super().__init__(*args, **kwargs)
        if score_mode not in self.supported_ops:
            raise AttributeError(
                'score_mode=%s is not supported! must be one of %s' % (score_mode, self.supported_ops.keys()))
        self.score_mode = score_mode

    @property
    def supported_ops(self):
        return {
            'multiply': np.prod,
            'sum': np.sum,
            'max': np.max,
            'min': np.min,
            'avg': np.mean,
        }

    def post_init(self):
        self.op = self.supported_ops[self.score_mode]

    def __call__(self, *last_scores, **kwargs) -> 'gnes_pb2.Response.QueryResponse.ScoredResult.Score':
        return self.new_score(
            value=self.op([s.value for s in last_scores]),
            operands=last_scores,
            score_mode=self.score_mode)


class ModifierScoreFn(BaseScoreFn):
    """Modifier to apply to the value
    score = modifier(factor * value)
    """

    def __init__(self, modifier: str = 'none', factor: float = 1.0, factor_name: str = 'GivenConstant',
                 *args, **kwargs):
        super().__init__(*args, **kwargs)
        if modifier not in self.supported_ops:
            raise AttributeError(
                'modifier=%s is not supported! must be one of %s' % (modifier, self.supported_ops.keys()))
        self._modifier = modifier
        self._factor = factor
        self._factor_name = factor_name

    @property
    def supported_ops(self):
        return {
            'none': lambda x: x,
            'log': np.log10,
            'log1p': lambda x: np.log10(x + 1),
            'log2p': lambda x: np.log10(x + 2),
            'ln': np.log,
            'ln1p': np.log1p,
            'ln2p': lambda x: np.log(x + 2),
            'square': np.square,
            'sqrt': np.sqrt,
            'reciprocal': np.reciprocal,
            'reciprocal1p': lambda x: np.reciprocal(1 + x),
            'abs': np.abs,
            'invert': lambda x: - x,
            'invert1p': lambda x: 1 - x
        }

    def post_init(self):
        self.factor = get_unary_score(value=self._factor, name=self._factor_name)
        self.op = self.supported_ops[self._modifier]

    def __call__(self,
                 last_score: 'gnes_pb2.Response.QueryResponse.ScoredResult.Score',
                 *args, **kwargs) -> \
            'gnes_pb2.Response.QueryResponse.ScoredResult.Score':
        if self._modifier == 'none' and self._factor == 1.0:
            return last_score
        else:
            return self.new_score(
                value=self.op(self.factor.value * last_score.value),
                operands=[last_score],
                modifier=self._modifier,
                factor=json.loads(self.factor.explained))


class ScoreOps:
    multiply = CombinedScoreFn('multiply')
    sum = CombinedScoreFn('sum')
    max = CombinedScoreFn('max')
    min = CombinedScoreFn('min')
    avg = CombinedScoreFn('avg')
    none = ModifierScoreFn('none')
    log = ModifierScoreFn('log')
    log1p = ModifierScoreFn('log1p')
    log2p = ModifierScoreFn('log2p')
    ln = ModifierScoreFn('ln')
    ln1p = ModifierScoreFn('ln1p')
    ln2p = ModifierScoreFn('ln2p')
    square = ModifierScoreFn('square')
    sqrt = ModifierScoreFn('sqrt')
    abs = ModifierScoreFn('abs')
    reciprocal = ModifierScoreFn('reciprocal')
    reciprocal1p = ModifierScoreFn('reciprocal1p')
